/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   mumps.c
 *    main module of freem
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdlib.h>
#include <stddef.h>
#include "mpsdef.h"
#include "errmsg.h"
#include "iniconf.h"
#include "namespace.h"
#include "transact.h"
#include "init.h"
#include <setjmp.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <limits.h>
#include <sys/types.h>
#include <pwd.h>
#include <string.h>

#include "version.h"

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)
# include <getopt.h>
#endif

#if !defined(PATH_MAX) && defined(_SCO_DS)
# define PATH_MAX 1024
#endif

extern int xecline(int typ);
extern char *getenv(const char *name);

int main (int argc, char **argv, char **envp)
{

    int c;
    int import_env;

    int option_index = 0;

#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)
    struct option long_options[] = {
        {"hardcopy", no_argument, &hardcopy, ENABLE},
        {"filter", no_argument, &filter, TRUE},
        {"noclear", no_argument, &noclear, TRUE},
        {"standard", no_argument, &standard, TRUE},
        {"import", no_argument, &import_env, TRUE},
        {"quiet", no_argument, &quiet_mode, TRUE},
        {"routine", required_argument, 0, 'r'},
        {"namespace", required_argument, 0, 'n'},
        {"config", required_argument, 0, 'c'},
        {0, 0, 0, 0}
    };
#endif
    char nsnbuf[256];
    
    char **env;      
    char *varname = (char *) calloc(STRLEN, sizeof(char));
    char *varval = (char *) calloc(STRLEN, sizeof(char));

    char *symname = (char *) calloc(STRLEN, sizeof(char));
    char *symval = (char *) calloc(STRLEN, sizeof(char));

    int namelen;
    int vallen;

    char cli_rtn_path[PATH_MAX];
    char *cli_rtn_file;
    char cli_rtn_name[256];

    m_argc = argc;			/* save arguments count     */
    m_argv = argv;			/* save arguments string    */
    m_envp = envp;			/* save environment pointer */

    strcpy (zb, argv[0]);		/* name with which mumps has been called */
    stcnv_c2m (zb);
    strcpy (stack0, argv[0]);
    stcnv_c2m (stack0);

    strncpy (config_file, SYSCONFDIR"/freem.conf", 100);
 
#if !defined(_SCO_DS) && !defined(__AMIGA) && !defined(_AIX)

    while(1) {
        
        c = getopt_long(argc, argv, "hfCsiqr:n:c:", long_options, &option_index);

        if (c == -1) break;

        switch(c) {
            case 'h':
                hardcopy = ENABLE;
                break;

            case 'f':
                filter = TRUE;
                break;

            case 'C':
                noclear = TRUE;
                break;

            case 's':
                standard = TRUE;
                break;

            case 'i':
                import_env = TRUE;
                break;

            case 'q':
                quiet_mode = TRUE;
                break;

            case 'c': /* specify FreeM configuration file */
                strncpy (config_file, optarg, 4095);
                break;

            case 'r': /* startup routine */
                direct_mode = FALSE;
                usermode = 0;

                strcpy (startuprou, optarg);
                startuprou[strlen (startuprou)] = '\201';
                break;

            case 'n': /* namespace */
                strcpy (nsname, optarg);               
                break;

        } 

    }

#else
    
    {
        extern char *optarg;
        extern int optind, optopt;

        while ((c = getopt(argc, argv, "hfCsiqr:n:c:")) != -1) {

            switch(c) {
                case 'h':
                    hardcopy = ENABLE;
                    break;

                case 'f':
                    filter = TRUE;
                    break;

                case 'C':
                    noclear = TRUE;
                    break;

                case 's':
                    standard = TRUE;
                    break;

                case 'i':
                    import_env = TRUE;
                    break;

                case 'q':
                    quiet_mode = TRUE;
                    break;

                case 'c':  /* specify FreeM configuration file */
                    strncpy (config_file, optarg, 4095);
                    break;

                case 'r': /* startup routine */
                    direct_mode = FALSE;
                    usermode = 0;

                    strcpy (startuprou, optarg);
                    startuprou[strlen (startuprou)] = '\201';
                    break;

                case 'n': /* namespace */
                    strcpy (nsname, optarg);               
                    break;

            } 
        }
    }
#endif

#if !defined(_AIX)
    if(import_env == TRUE) {

        for(env = envp; *env != 0; env++) {
            
            namelen = 0;
            vallen = 0;
            
            varname = strtok(*env, "=");
            varval = strtok(NULL, "=");

            if(varval != NULL) {
                namelen = strlen(varname);
                vallen = strlen(varval);

                strncpy(symname, varname, STRLEN);
                strncpy(symval, varval, STRLEN);

                stcnv_m2c (symname);
                stcnv_m2c (symval);

                symtab(set_sym, symname, symval);
            }
        }
    }
#endif    

    /* handle passing of an arbitrary .m file on the command line */
    /* this is most often used for shebang-line scripts. */
    if (optind < argc) {

        /* bail if file does not exist */
        if (access (argv[optind], F_OK) == -1) {

            set_io (UNIX);
            
            fprintf (stderr, "Routine %s does not exist.\n", argv[optind]);

            exit (1);

        }

        direct_mode = FALSE;
        usermode = 0;

 

        /* was a path specified at all? */
        if (strchr (argv[optind], '/') == NULL) {

            /* the entirety of argv[optind] is the filename */
            cli_rtn_file = argv[optind];            

            /* use the current directory */
            sprintf (cli_rtn_path, ".");

        }
        else {

            /* isolate the filename from the path */
            cli_rtn_file = strrchr (argv[optind], '/') + 1;
            
            /* isolate the routine name from the filename */
            strncpy (cli_rtn_name, cli_rtn_file, strchr (cli_rtn_file, '.') - cli_rtn_file);
            
            /* isolate the path from the routine file */
            strncpy (cli_rtn_path, argv[optind], strrchr (argv[optind], '/') - argv[optind]);

        }

        /* do we have a file extension? */
        if (strchr (cli_rtn_file, '.') != NULL) {
            
            /* if so, just remove it */
            strncpy (cli_rtn_name, cli_rtn_file, strchr (cli_rtn_file, '.') - cli_rtn_file);

        }
        else {

            /* otherwise, just take a direct copy */
            strcpy (cli_rtn_name, cli_rtn_file);

        }

        /* make this the startup routine */
        snprintf (startuprou, 256, "^%s\201", cli_rtn_name);

        /* re-work the namespace config to search for the 
           routine in the discovered path */
        if (cli_rtn_name[0] == '%') {

            snprintf (rou0plib, 256, "%s\201", cli_rtn_path);
            snprintf (rou1plib, 256, "%s\201", cli_rtn_path);

        }
        else {

            snprintf (rou0path, 256, "%s\201", cli_rtn_path);
            snprintf (rou1path, 256, "%s\201", cli_rtn_path);

        }

    }

    strncpy (nsnbuf, nsname, 255);    

#ifdef __EMSCRIPTEN__
    fprintf (stderr, "EMSCRIPTEN CHECK PRE-INIT\n");
#endif

    /* initialize FreeM environment */
    if (init (nsnbuf) == FALSE) {

        set_io (UNIX);
        fprintf (stderr, "\nError initializing FreeM.\n");

        exit (1);

    }

    if (direct_mode == TRUE && quiet_mode == FALSE) {
        char version[256];
        
        snprintf (version, 256, "Coherent Logic Development FreeM\r\nVersion %s\r\n\r\n\201", FREEM_VERSION_ID);
        write_m (version);

        write_m ("           *\r\n\201");
        write_m ("          * *\r\n\201");
        write_m ("         *   *\r\n\201");
        write_m ("    ***************\r\n\201");
        write_m ("     * *       * *\r\n\201");
        write_m ("      *  FreeM  *\r\n\201");
        write_m ("     * *       * *\r\n\201");
        write_m ("    ***************\r\n\201");
        write_m ("         *   *\r\n\201");
        write_m ("          * *     Copyright (C) 1998 MUG Deutschland\r\n\201");
        write_m ("           *      Copyright (C) 2014, 2020 Coherent Logic Development LLC\r\n\r\n\201");

        write_m ("This program is free software: you can redistribute it and/or modify\r\n\201");
        write_m ("it under the terms of the GNU Affero General Public License as published by\r\n\201");
        write_m ("the Free Software Foundation, either version 3 of the License, or\r\n\201");
        write_m ("(at your option) any later version.\r\n\r\n\201");
        
        write_m ("This program is distributed in the hope that it will be useful,\r\n\201");
        write_m ("but WITHOUT ANY WARRANTY; without even the implied warranty of\r\n\201");
        write_m ("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\r\n\201");
        write_m ("GNU Affero General Public License for more details.\r\n\r\n\201");

        write_m ("You should have received a copy of the GNU Affero General Public License\r\n\201");
        write_m ("along with this program.  If not, see <https://www.gnu.org/licenses/>.\r\n\r\n\201");

    }
    else {
        write_m ("\r\n\r\n\201");
    }

#ifndef __EMSCRIPTEN__
    /* run mumps */
    xecline (1);
#else
    fprintf (stderr, "EMSCRIPTEN CHECK PRE-XECLINE\n");
    stcpy (code, "QUIT\201");
    xecline (4);
#endif

    exit (0);	/* we should never reach that statement */

} /* end of main() */

